import pytest
from tests.conftest import standardizedEqualTo
from updateCitation import addPyPAMetadata, CitationNexus
from updateCitation.pypa import getPyPAMetadata

def test_getPyPAMetadata_missingName() -> None:
	dictionaryPackageData = {
		"version": "17.19.23",
	}
	with pytest.raises(Exception):
		getPyPAMetadata(dictionaryPackageData)



from updateCitation.variables import (
	CitationNexus,
	CitationNexusFieldsProtected,
	filename_pyprojectDOTtomlDEFAULT,
	formatDateCFF,
	FREAKOUT,
	gitUserEmailFALLBACK,
	mapNexusCitation2pyprojectDOTtoml,
	SettingsPackage,
	Z0Z_mappingFieldsURLFromPyPAMetadataToCFF,
)

from updateCitation.pyprojectDOTtoml import add_pyprojectDOTtoml, getSettingsPackage
from updateCitation.citationFileFormat import addCitation, writeCitation
from updateCitation.pypa import addPyPAMetadata, compareVersions
from updateCitation.github import addGitHubRelease, addGitHubSettings, gittyUpGitAmendGitHub
from updateCitation.pypi import addPyPIrelease

from updateCitation.flowControl import here as here



from pathlib import Path
from updateCitation import CitationNexus, SettingsPackage
from updateCitation.variables import cffDASHversionDefaultHARDCODED, messageDefaultHARDCODED

def test_CitationNexus_requiredFields(nexusCitationTesting: CitationNexus) -> None:
	assert nexusCitationTesting.cffDASHversion == cffDASHversionDefaultHARDCODED
	assert nexusCitationTesting.message == messageDefaultHARDCODED
	assert nexusCitationTesting.authors == []
	assert nexusCitationTesting.title is None

def test_SettingsPackage_initialization(pathFilenameTmpTesting: Path) -> None:
	settings = SettingsPackage(pathFilenamePackageSSOT=pathFilenameTmpTesting)
	assert settings.pathFilenamePackageSSOT == pathFilenameTmpTesting
	assert settings.filenameCitationDOTcff == "CITATION.cff"
	assert isinstance(settings.tomlPackageData, dict)



"""SSOT for Pytest."""
from collections.abc import Callable, Generator
from pathlib import Path
from typing import Any
from updateCitation import (
	CitationNexus, CitationNexusFieldsProtected, filename_pyprojectDOTtomlDEFAULT, SettingsPackage,
)
import pytest
import shutil
import uuid

# SSOT for test data paths and filenames
pathDataSamples = Path("tests/dataSamples")
pathFilenameCitationAlphaDOTcff = pathDataSamples / "citationAlpha.cff"
pathTmpRoot = pathDataSamples / "tmp"

@pytest.fixture(scope="session", autouse=True)
def setupTeardownTestSession() -> Generator[None]:
	pathDataSamples.mkdir(exist_ok=True)
	pathTmpRoot.mkdir(exist_ok=True)
	yield
	shutil.rmtree(pathTmpRoot, ignore_errors=True)

@pytest.fixture(autouse=True)
def resetCitationNexusProtectedFields() -> None:
	"""Reset the protected fields set before each test to prevent test interference."""
	CitationNexusFieldsProtected.clear()

@pytest.fixture
def pathTmpTesting(request: pytest.FixtureRequest) -> Path:
	"""'path' means directory or folder, not file."""
	pathTmp = pathTmpRoot / uuid.uuid4().hex
	pathTmp.mkdir(parents=True, exist_ok=False)
	return pathTmp

@pytest.fixture
def pathFilenameTmpTesting(request: pytest.FixtureRequest) -> Path:
	"""'filename' means file; 'pathFilename' means the full path and filename."""
	try:
		extension = request.param
	except AttributeError:
		extension = ".txt"

	uuidString = uuid.uuid4().hex
	pathFilenameTmp = Path(pathTmpRoot, uuidString[0:-8], uuidString[-8:None] + extension)
	pathFilenameTmp.parent.mkdir(parents=True, exist_ok=False)
	return pathFilenameTmp

@pytest.fixture
def nexusCitationTesting(request: pytest.FixtureRequest) -> CitationNexus:
	"""Return a CitationNexus object with the specified attributes."""
	try:
		attributes: dict[str, Any] = request.param
	except AttributeError:
		attributes = {}

	return CitationNexus(**attributes)

@pytest.fixture
def citationAlphaDOTcff() -> Path:
	return pathFilenameCitationAlphaDOTcff

"""
Section: Pytest fixtures for testing the updateCitation package"""

@pytest.fixture
def settingsPackageTesting() -> SettingsPackage:
	return SettingsPackage(
		pathFilenamePackageSSOT= pathDataSamples / filename_pyprojectDOTtomlDEFAULT,
		GITHUB_TOKEN="FAKE_TOKEN",
		gitUserEmail="test@example.com",
		gitUserName="TestUserName",
		gitCommitMessage="TestCommitMessage",
	)

"""
Section: Standardized assert statements and failure messages"""

def uniformTestFailureMessage(expected: Any, actual: Any, functionName: str, *arguments: Any, **keywordArguments: Any) -> str:
	"""Format assertion message for any test comparison."""
	listArgumentComponents = [str(parameter) for parameter in arguments]
	listKeywordComponents = [f"{key}={value}" for key, value in keywordArguments.items()]
	joinedArguments = ', '.join(listArgumentComponents + listKeywordComponents)

	return (f"\nTesting: `{functionName}({joinedArguments})`\n"
			f"Expected: {expected}\n"
			f"Got: {actual}")

def standardizedEqualTo(expected: Any, functionTarget: Callable[..., Any], *arguments: Any, **keywordArguments: Any) -> None:
	"""Template for most tests to compare the actual outcome with the expected outcome, including expected errors."""
	if type(expected) == type[Exception]:  # noqa: E721
		messageExpected = expected.__name__
	else:
		messageExpected = expected

	try:
		messageActual = actual = functionTarget(*arguments, **keywordArguments)
	except Exception as actualError:
		messageActual = type(actualError).__name__
		actual = type(actualError)

	assert actual == expected, uniformTestFailureMessage(messageExpected, messageActual, functionTarget.__name__, *arguments, **keywordArguments)



from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from updateCitation import (
	CitationNexus, compareVersions, formatDateCFF, FREAKOUT, gitUserEmailFALLBACK, SettingsPackage)
import datetime
import github
import github.Repository
import os
import warnings

@contextmanager
def GithubClient(tokenAsStr: str | None) -> Generator[github.Github, Any, None]:
	"""Creates a GitHub client with authentication if a token is available.

	Parameters:
		tokenAsStr: A GitHub authentication token as a string.

	Returns:
		A GitHub client instance that will be automatically closed.
	"""
	# Don't do the following because: 1) not DRY, 2) it's "sneaky" to use environment variables without the user's knowledge.
	# tokenAsStr = tokenAsStr or os.environ.get("GITHUB_TOKEN") or os.environ.get("GH_TOKEN")
	githubAuthToken = github.Auth.Token(tokenAsStr) if tokenAsStr else None
	githubClient = github.Github(auth=githubAuthToken)
	try:
		yield githubClient
	finally:
		githubClient.close()

def addGitHubSettings(truth: SettingsPackage) -> SettingsPackage:
	# TODO low priority: make "load token from environment variable" optional
	truth.GITHUB_TOKEN = truth.GITHUB_TOKEN or os.environ.get("GITHUB_TOKEN") or os.environ.get("GH_TOKEN")

	if not truth.gitUserEmail:
		with GithubClient(truth.GITHUB_TOKEN) as githubClient:
			try:
				userGitHub = githubClient.get_user()
				ImaGitUserEmail = f"{userGitHub.id}+{userGitHub.login}@users.noreply.github.com"
			except github.GithubException:
				ImaGitUserEmail = None

		githubActor = os.environ.get("GITHUB_ACTOR")
		gitUserEmailGithubActor = f"{githubActor}@users.noreply.github.com" if githubActor else None
		truth.gitUserEmail = ImaGitUserEmail or gitUserEmailGithubActor or gitUserEmailFALLBACK

	return truth

@contextmanager
def GitHubRepository(nexusCitation: CitationNexus, truth: SettingsPackage) -> Generator[github.Repository.Repository, Any, None]:
	"""Creates a GitHub repository instance for the given citation.

	Parameters:
		nexusCitation: Citation object containing repository information.
		truth: Settings package containing GitHub authentication token.

	Returns:
		A GitHub repository instance.

	Raises:
		FREAKOUT: If repository information is missing.
	"""
	if not nexusCitation.repository:
		raise FREAKOUT

	with GithubClient(truth.GITHUB_TOKEN) as githubClient:
		full_name_or_id: str = nexusCitation.repository.replace("https://github.com/", "").replace(".git", "")
		githubRepository = githubClient.get_repo(full_name_or_id)
		yield githubRepository

def gittyUpGitAmendGitHub(truth: SettingsPackage, nexusCitation: CitationNexus, pathFilenameCitationSSOT: Path, pathFilenameCitationDOTcffRepository: Path):
	environmentIsGitHubAction = bool(os.environ.get("GITHUB_ACTIONS") and os.environ.get("GITHUB_WORKFLOW"))
	if not environmentIsGitHubAction or not nexusCitation.repository:
		return

	import subprocess

	# TODO I don't like that this flow assumes `git` is installed and available in the environment.
	# Can I use `GitHubRepository` instead of `subprocess`?

	subprocess.run(["git", "config", "user.name", truth.gitUserName])
	subprocess.run(["git", "config", "user.email", truth.gitUserEmail])
	# Get the previous commit message
	previousCommitResult = subprocess.run(["git", "log", "-1", "--pretty=format:%s"], capture_output=True, text=True)
	if previousCommitResult.returncode == 0 and previousCommitResult.stdout.strip():
		previousCommitMessage = previousCommitResult.stdout.strip()
		# Only append if the previous message doesn't already contain citation update text
		if "Update CITATION.cff" not in previousCommitMessage:
			combinedCommitMessage = f"{previousCommitMessage} + Update CITATION.cff [skip ci]"
		else:
			combinedCommitMessage = truth.gitCommitMessage
	else:
		combinedCommitMessage = truth.gitCommitMessage

	# Stage the citation files
	subprocess.run(["git", "add", str(pathFilenameCitationSSOT), str(pathFilenameCitationDOTcffRepository)])

	commitResult = subprocess.run(["git", "commit", "-m", combinedCommitMessage])
	if commitResult.returncode == 0:
		subprocess.run(["git", "push", "origin", "HEAD"])

def getGitHubRelease(nexusCitation: CitationNexus, truth: SettingsPackage) -> dict[str, Any] | None:
	"""Retrieves the latest release information from a GitHub repository.
		Parameters:
			nexusCitation: A CitationNexus object containing citation metadata, including the repository URL.
		Returns:
			dictionaryRelease: A dictionary containing release information or an empty dictionary.
		Raises:
			NEVER: `Exception` is caught and converted to a warning. (So, don't filter all warnings, you know?)
		"""
	if not nexusCitation.repository:
		return None

	try:
		# NOTE latestRelease.tag_name == nexusCitation.version
		if not nexusCitation.version:
			raise FREAKOUT

		with GitHubRepository(nexusCitation, truth) as githubRepository:
			latestRelease = githubRepository.get_latest_release()
			tagObject = githubRepository.get_git_ref(f'tags/{latestRelease.tag_name}').object
			# TODO `commitLatestRelease` should be fixed but it's not
			commitLatestRelease = tagObject.sha if tagObject.type == 'tag' else tagObject.sha
			commitLatestCommit = githubRepository.get_commit(githubRepository.default_branch).sha

		urlRelease: str = latestRelease.html_url

		dictionaryRelease: dict[str, Any] = {
			"commit": commitLatestRelease,
			"dateDASHreleased": latestRelease.published_at.strftime(formatDateCFF),
			"identifiers": [{
				"type": "url",
				"value": urlRelease,
				"description": f"The URL for {nexusCitation.title} {latestRelease.tag_name}."
			}] if urlRelease else [],
			"repositoryDASHcode": urlRelease,
		}

		if compareVersions(latestRelease.tag_name, nexusCitation.version) == -1:
			dictionaryReleaseHypothetical: dict[str, Any] = {
				"commit": commitLatestCommit,
				"dateDASHreleased": datetime.datetime.now().strftime(formatDateCFF),
				"identifiers": [{
					"type": "url",
					"value": urlRelease.replace(latestRelease.tag_name, nexusCitation.version),
					"description": f"The URL for {nexusCitation.title} {nexusCitation.version}."
				}] if urlRelease else [],
				"repositoryDASHcode": urlRelease.replace(latestRelease.tag_name, nexusCitation.version),
			}
			dictionaryRelease.update(dictionaryReleaseHypothetical)

		return dictionaryRelease

	except Exception as ERRORmessage:
		warnings.warn(f"Failed to get GitHub release information. {ERRORmessage}", UserWarning)
		return None

def addGitHubRelease(nexusCitation: CitationNexus, truth: SettingsPackage) -> CitationNexus:
	"""Adds GitHub release information to a CitationNexus object.
		Parameters:
			nexusCitation: The CitationNexus object to update.
		Returns:
			CitationNexus: The updated CitationNexus object with GitHub release information.
	"""

	gitHubReleaseData = getGitHubRelease(nexusCitation, truth)

	if gitHubReleaseData:
		commitSherpa = gitHubReleaseData.get("commit")
		if commitSherpa:
			nexusCitation.commit = commitSherpa

		dateDASHreleasedSherpa = gitHubReleaseData.get("dateDASHreleased")
		if dateDASHreleasedSherpa:
			nexusCitation.dateDASHreleased = dateDASHreleasedSherpa

		identifiersSherpa = gitHubReleaseData.get("identifiers")
		if identifiersSherpa:
			nexusCitation.identifiers = identifiersSherpa

		repositoryDASHcodeSherpa = gitHubReleaseData.get("repositoryDASHcode")
		if repositoryDASHcodeSherpa:
			nexusCitation.repositoryDASHcode = repositoryDASHcodeSherpa

	nexusCitation.setInStone("GitHub")
	return nexusCitation



from packaging.metadata import Metadata as PyPAMetadata
from typing import Any, cast
from updateCitation import CitationNexus, Z0Z_mappingFieldsURLFromPyPAMetadataToCFF
import packaging
import packaging.metadata
import packaging.utils
import packaging.version

def compareVersions(comparand: str, comparator: str) -> int:
	"""
	Compares two version strings using packaging.version.Version.
	Parameters:
		comparand: The version string to compare.
		comparator: The version string to compare against.
	Returns:
		relationship: '-1' if `comparand` is less than `comparator`, '0' if they are equal, '1' if `comparand` is greater than `comparator`, or 'else'.
	"""
	versionComparand = packaging.version.Version(comparand)
	versionComparator = packaging.version.Version(comparator)
	if versionComparand < versionComparator:
		return -1
	elif versionComparand > versionComparator:
		return 1
	elif versionComparand == versionComparator:
		return 0
	else:
		return 3153

def getPyPAMetadata(packageData: dict[str, Any]) -> PyPAMetadata:
	"""
	Retrieves and formats package metadata from a dictionary into a PyPAMetadata object.
	Parameters:
		packageData: A dictionary containing package information.
	Returns:
		PyPAMetadata: A PyPAMetadata object containing the extracted and formatted
			metadata. The package name is canonicalized using `packaging.utils.canonicalize_name` and validated.
	"""
	dictionaryPackageDataURLs: dict[str, str] = packageData.get("urls", {})
	dictionaryProjectURLs: dict[str, str] = {}
	for urlName, url in dictionaryPackageDataURLs.items():
		urlName = urlName.lower()
		dictionaryProjectURLs[urlName] = url

	dictionaryPackageDataLicense: dict[str, str] = packageData.get("license", {})
	metadataRaw = packaging.metadata.RawMetadata(
		keywords=packageData.get("keywords", []),
		license_expression=dictionaryPackageDataLicense.get("text", ""),
		metadata_version="2.4",
		# NOTE packaging.metadata.InvalidMetadata: 'name' is a required field
		name=cast(str, packaging.utils.canonicalize_name(packageData.get("name", None), validate=True)), # pyright: ignore[reportArgumentType]
		project_urls=dictionaryProjectURLs,
		version=cast(str, packageData.get("version", None)),  # pyright: ignore[reportArgumentType]
	)

	metadata = PyPAMetadata().from_raw(metadataRaw)
	return metadata

def addPyPAMetadata(nexusCitation: CitationNexus, tomlPackageData: dict[str, Any], projectURLTargets: set[str]) -> CitationNexus:
	pypaMetadata: PyPAMetadata = getPyPAMetadata(tomlPackageData)

	if pypaMetadata.version:
		nexusCitation.version = str(pypaMetadata.version)
	if pypaMetadata.keywords:
		nexusCitation.keywords = pypaMetadata.keywords
	if pypaMetadata.license_expression:
		nexusCitation.license = pypaMetadata.license_expression

	if pypaMetadata.project_urls:
		for urlTarget in projectURLTargets:
			url = pypaMetadata.project_urls.get(urlTarget, None)
			if url:
				setattr(nexusCitation, Z0Z_mappingFieldsURLFromPyPAMetadataToCFF[urlTarget], url)

	nexusCitation.setInStone("PyPA")
	return nexusCitation



import pathlib
import pytest
from tests.conftest import standardizedEqualTo
from updateCitation.citationFileFormat import addCitation
from updateCitation.variables import CitationNexus

def test_addCitation(nexusCitationTesting: CitationNexus, citationAlphaDOTcff: pathlib.Path):
	"""Test that addCitation() returns a CitationNexus object."""
	nexusCitation = addCitation(nexusCitationTesting, citationAlphaDOTcff)
	assert isinstance(nexusCitation, CitationNexus), "addCitation() should return a CitationNexus object."

@pytest.mark.parametrize("fieldName, expectedValue", [
	("cffDASHversion", "1.2.0"),
	("message", "If you use this software, you can cite it using the metadata from this file."),
	("authors", [{'given-names': 'Hunter', 'family-names': 'Hogan', 'email': 'HunterHogan@pm.me'}]),
	("commit", "0d70eb67ffbab9563208ec06294887fb7fa6768a"),
	("dateDASHreleased", "2025-02-07"),
	("identifiers", [{'type': 'url', 'value': 'https://github.com/hunterhogan/mapFolding/releases/tag/0.3.9', 'description': 'The URL for mapFolding 0.3.9.'}]),
	("keywords", ['A001415', 'A001416', 'A001417', 'A001418', 'A195646', 'folding', 'map folding', 'OEIS', 'stamp folding']),
	("license", "CC-BY-NC-4.0"),
	("repository", "https://github.com/hunterhogan/mapFolding.git"),
	("repositoryDASHartifact", "https://pypi.org/project/mapfolding/0.3.9/"),
	("repositoryDASHcode", "https://github.com/hunterhogan/mapFolding/releases/tag/0.3.9"),
	("title", "mapFolding"),
	("type", "software"),
	("url", "https://github.com/hunterhogan/mapFolding"),
	("version", "0.3.9"),
])
def test_addCitation_fields(nexusCitationTesting: CitationNexus, citationAlphaDOTcff: pathlib.Path, fieldName: str, expectedValue: str):
	"""Test that addCitation() correctly sets each field in the CitationNexus object."""
	nexusCitation = addCitation(nexusCitationTesting, citationAlphaDOTcff)
	# actual = getattr(nexusCitation, fieldName)
	standardizedEqualTo(expectedValue, getattr, nexusCitation, fieldName)



from typing import Any, cast, Literal, TypedDict
import attrs
import inspect
import pathlib
import warnings

"""
Long term:
	`fieldsSSOT` will be something more like, the SSOT for this field is: ____
	`Z0Z_addGitHubRelease` will be unnecessary. The flow will cycle through the SSOTs for each field. If the SSOT for a field is GitHub, then the flow will add the GitHub release.
"""

# TODO think of a clever way to dynamically set the default version
cffDASHversionDefaultHARDCODED: str = '1.2.0'
# TODO change this to dynamically load the schema default message
messageDefaultHARDCODED: str = "Cite this software with the metadata in this file."
# TODO dynamically load through the following:
CitationNexusFieldsRequiredHARDCODED: set[str] = {"authors", "cffDASHversion", "message", "title"}
"""
from cffconvert.citation import Citation # from cffconvert.lib.citation import Citation # upcoming version 3.0.0
cffstr = "cff-version: 1.2.0"; citationObject = Citation(cffstr); schemaDOTjson = citationObject._get_schema()
# get "required": list of fields; # Convert '-' to 'DASH' in field names """

filename_pyprojectDOTtomlDEFAULT: str = 'pyproject.toml' # used by other processes before `SettingsPackage` is instantiated to help instantiate `SettingsPackage`
formatDateCFF: str = "%Y-%m-%d"
gitUserEmailFALLBACK: str = 'action@github.com'
mapNexusCitation2pyprojectDOTtoml: list[tuple[str, str]] = [("authors", "authors"), ("contact", "maintainers")]
Z0Z_mappingFieldsURLFromPyPAMetadataToCFF: dict[str, str] = {
	"homepage": "url",
	"license": "licenseDASHurl",
	"repository": "repository",
}

class FREAKOUT(Exception):
	pass

@attrs.define(slots=False)
class SettingsPackage:
	pathRepository: pathlib.Path = pathlib.Path.cwd()
	filename_pyprojectDOTtoml: str = filename_pyprojectDOTtomlDEFAULT
	pathFilenamePackageSSOT: pathlib.Path = pathlib.Path(pathRepository, filename_pyprojectDOTtoml)

	filenameCitationDOTcff: str = 'CITATION.cff'
	pathFilenameCitationDOTcffRepository: pathlib.Path = pathlib.Path(pathRepository, filenameCitationDOTcff)
	pathFilenameCitationSSOT: pathlib.Path = pathlib.Path(pathFilenameCitationDOTcffRepository)

	Z0Z_addGitHubRelease: bool = True
	Z0Z_addPyPIrelease: bool = True

	pathReferences: pathlib.Path = pathlib.Path(pathRepository, 'citations')
	projectURLTargets: set[str] = {"homepage", "license", "repository"}

	gitCommitMessage: str = "Update citations [skip ci]"
	gitUserName: str = "updateCitation"
	gitUserEmail: str = ""
	gitAmendFromGitHubAction: bool = True
	# gitPushFromOtherEnvironments_why_where_NotImplemented: bool = False
	tomlPackageData: dict[str, Any] = cast(dict[str, Any], attrs.field(factory=dict))

	GITHUB_TOKEN: str | None = None

CitationNexusFieldsRequired: set[str] = CitationNexusFieldsRequiredHARDCODED
CitationNexusFieldsProtected: set[str] = set()

# Define type definitions for schema structures
class Person(TypedDict, total=False):
	address: str
	affiliation: str
	alias: str
	city: str
	country: str
	email: str
	family_names: str
	fax: str
	given_names: str
	name_particle: str
	name_suffix: str
	orcid: str
	post_code: str | int
	region: str
	tel: str
	website: str

class Entity(TypedDict, total=False):
	address: str
	alias: str
	city: str
	country: str
	date_end: str
	date_start: str
	email: str
	fax: str
	location: str
	name: str
	orcid: str
	post_code: str | int
	region: str
	tel: str
	website: str

class Identifier(TypedDict, total=False):
	description: str
	type: Literal["doi", "url", "swh", "other"]
	value: str

class ReferenceDictionary(TypedDict, total=False):
	abbreviation: str
	abstract: str
	authors: list[Person | Entity]
	collection_doi: str
	collection_title: str
	collection_type: str
	commit: str
	conference: Entity
	contact: list[Person | Entity]
	copyright: str
	data_type: str
	database: str
	database_provider: Entity
	date_accessed: str
	date_downloaded: str
	date_published: str
	date_released: str
	department: str
	doi: str
	edition: str
	editors: list[Person | Entity]
	editors_series: list[Person | Entity]
	end: int | str
	entry: str
	filename: str
	format: str
	identifiers: list[Identifier]
	institution: Entity
	isbn: str
	issn: str
	issue: str | int
	issue_date: str
	issue_title: str
	journal: str
	keywords: list[str]
	languages: list[str]
	license: str | list[str]
	license_url: str
	loc_end: int | str
	loc_start: int | str
	location: Entity
	medium: str
	month: int | str
	nihmsid: str
	notes: str
	number: str | int
	number_volumes: int | str
	pages: int | str
	patent_states: list[str]
	pmcid: str
	publisher: Entity
	recipients: list[Entity | Person]
	repository: str
	repository_artifact: str
	repository_code: str
	scope: str
	section: str | int
	senders: list[Entity | Person]
	start: int | str
	status: Literal["abstract", "advance-online", "in-preparation", "in-press", "preprint", "submitted"]
	term: str
	thesis_type: str
	title: str
	translators: list[Entity | Person]
	type: str
	url: str
	version: str | int
	volume: int | str
	volume_title: str
	year: int | str
	year_original: int | str

@attrs.define()
class CitationNexus:
	"""one-to-one correlation with `cffconvert.lib.cff_1_2_x.citation` class Citation_1_2_x.cffobj"""
	abstract: str | None = None
	authors: list[dict[str, str]] = cast(list[dict[str, str]], attrs.field(factory=list))
	cffDASHversion: str = cffDASHversionDefaultHARDCODED
	commit: str | None = None
	contact: list[dict[str, str]] = cast(list[dict[str, str]], attrs.field(factory=list))
	dateDASHreleased: str | None = None
	doi: str | None = None
	identifiers: list[str] = cast(list[str], attrs.field(factory=list))
	keywords: list[str] = cast(list[str], attrs.field(factory=list))
	license: str | None = None
	licenseDASHurl: str | None = None
	message: str = messageDefaultHARDCODED
	preferredDASHcitation: ReferenceDictionary | None = None
	# TODO `cffconvert` also doesn't convert references yet
	references: list[ReferenceDictionary] = cast(list[ReferenceDictionary], attrs.field(factory=list))
	repository: str | None = None
	repositoryDASHartifact: str | None = None
	repositoryDASHcode: str | None = None
	title: str | None = None
	type: str | None = None
	url: str | None = None
	version: str | None = None

	# NOTE the names of the existing parameters for `__setattr__` are fixed
	def __setattr__(self, name: str, value: Any, warn: bool | None = True) -> None:
		"""Prevent modification of protected fields."""
		if name in CitationNexusFieldsProtected:
			if warn:
				# Get the line of code that called this method
				stackFrame = inspect.stack()[1]
				codeContext = stackFrame.code_context[0] if stackFrame.code_context is not None else ""
				context = codeContext.strip()
				# TODO Improve this warning message and the context information.
				warnings.warn(f"A process tried to change the field '{name}' after the authoritative source set the field's value.\n{context=}", UserWarning)
			return
		super().__setattr__(name, value)

	def setInStone(self, prophet: str) -> None:
		"""
		Confirm that required fields are not None, and freeze fields specified by the context.
		Parameters:
			prophet: The power to protect a field.
		Returns:
			None:
		Raises:
			ValueError: A required field does not have a value.
		"""
		match prophet:
			case "Citation":
				fieldsSSOT: set[str] = {"abstract", "cffDASHversion", "doi", "message", "preferredDASHcitation", "type"}
			case "GitHub":
				fieldsSSOT = {"commit", "dateDASHreleased", "identifiers", "repositoryDASHcode"}
			case "PyPA":
				fieldsSSOT = {"keywords", "license", "licenseDASHurl", "repository", "url", "version"}
			case "PyPI":
				fieldsSSOT = {"repositoryDASHartifact"}
			case "pyprojectDOTtoml":
				fieldsSSOT = {"authors", "contact", "title"}
			case _:
				fieldsSSOT = set()

		for fieldName in fieldsSSOT:
			if fieldName in CitationNexusFieldsRequired and not getattr(self, fieldName, None):
				# TODO work out the semiotics of SSOT, power, authority, then improve this message (and identifiers and your life and the world)
				raise ValueError(f"I have not yet received a value for the field '{fieldName}', but the Citation Field Format requires the field and {prophet} should have provided it.")

		CitationNexusFieldsProtected.update(fieldsSSOT)



from os import PathLike
from pathlib import Path
from typing import Any
from updateCitation import (
	add_pyprojectDOTtoml, addCitation, addGitHubRelease, addGitHubSettings, addPyPAMetadata, addPyPIrelease, CitationNexus,
	filename_pyprojectDOTtomlDEFAULT, getSettingsPackage, gittyUpGitAmendGitHub, SettingsPackage, writeCitation)

def here(pathFilename_pyprojectDOTtoml: str | PathLike[Any] | None = None) -> None:
	pathFilenameSettingsSSOT = Path(pathFilename_pyprojectDOTtoml) if pathFilename_pyprojectDOTtoml else Path.cwd() / filename_pyprojectDOTtomlDEFAULT
	truth: SettingsPackage = getSettingsPackage(pathFilenameSettingsSSOT)

	nexusCitation = CitationNexus()

	nexusCitation = add_pyprojectDOTtoml(nexusCitation, truth)

	if not nexusCitation.title:
		# TODO learn how to change the field from `str | None` to `str` after the field is populated
		# especially for a required field
		raise ValueError("Package name is required.")

	if Path(truth.pathFilenameCitationSSOT).exists():
		pathFilenameCitationSSOT = truth.pathFilenameCitationSSOT
	elif Path(truth.pathFilenameCitationDOTcffRepository).exists():
		pathFilenameCitationSSOT = truth.pathFilenameCitationDOTcffRepository
	else:
		truth.pathFilenameCitationSSOT.parent.mkdir(parents=True, exist_ok=True)
		truth.pathFilenameCitationSSOT.write_text(f"cff-version: {nexusCitation.cffDASHversion}\n")
		pathFilenameCitationSSOT = truth.pathFilenameCitationSSOT

	nexusCitation = addCitation(nexusCitation, pathFilenameCitationSSOT)
	nexusCitation = addPyPAMetadata(nexusCitation, truth.tomlPackageData, truth.projectURLTargets)
	truth = addGitHubSettings(truth)
	if truth.Z0Z_addGitHubRelease:
		nexusCitation = addGitHubRelease(nexusCitation, truth)
	if truth.Z0Z_addPyPIrelease:
		nexusCitation = addPyPIrelease(nexusCitation)

	validationStatus = writeCitation(nexusCitation, truth.pathFilenameCitationSSOT, truth.pathFilenameCitationDOTcffRepository)

	"""TODO remove the second push
	TODO figure out the sha paradox
	TODO possibly related: fix the `commitLatestRelease` value in `getGitHubRelease`
	- allegedly, `commitInProgress = os.environ.get("GITHUB_SHA")`
	- so the citation could 1) have the correct commit hash in the same file as the release,
	and 2) the up-to-date citation file could be in the release it references.

	During some commits, I intentionally make both `dictionaryRelease` and `dictionaryReleaseHypothetical`.
	But I don't know how to conditionally use `dictionaryReleaseHypothetical` only if the Python Tests pass
	(and the release actions are successful).

	I guess I could wait to see the outcome of the tests,
	then choose the correct dictionary. I don't want to prevent the commit: I just want to put accurate information
	in the citation file.

	"""

	if validationStatus and truth.gitAmendFromGitHubAction:
		gittyUpGitAmendGitHub(truth, nexusCitation, truth.pathFilenameCitationSSOT, truth.pathFilenameCitationDOTcffRepository)



from cffconvert.cli.create_citation import create_citation
from typing import Any
from updateCitation import CitationNexus
import attrs
import cffconvert
import pathlib
import ruamel.yaml

def getCitation(pathFilenameCitationSSOT: pathlib.Path) -> dict[str, Any]:
	# Try to converge with cffconvert when possible.
	citationObject: cffconvert.Citation = create_citation(infile=str(pathFilenameCitationSSOT), url=None)
	# `._parse()` is `ruamel.yaml.YAML` loader
	return citationObject._parse()

def addCitation(nexusCitation: CitationNexus, pathFilenameCitationSSOT: pathlib.Path) -> CitationNexus:
	cffobj = getCitation(pathFilenameCitationSSOT)

	# This step is designed to prevent deleting fields that are populated in the current CFF file,
	# but for whatever reason do not get added to the CitationNexus object.

	for nexusCitationField in iter(attrs.fields(type(nexusCitation))):
		cffobjKeyName: str = nexusCitationField.name.replace("DASH", "-")
		cffobjValue = cffobj.get(cffobjKeyName)
		if cffobjValue: # An empty list will be False
			nexusCitation.__setattr__(nexusCitationField.name, cffobjValue, warn=False)

	nexusCitation.setInStone("Citation")
	return nexusCitation

def writeCitation(nexusCitation: CitationNexus, pathFilenameCitationSSOT: pathlib.Path, pathFilenameCitationDOTcffRepo: pathlib.Path | None = None) -> bool:
	# NOTE embarrassingly hacky process to follow

	# TODO format the output
	# parameterIndent: int = 2
	# parameterLineWidth: int = 60

	# use `ruamel.yaml` because using the same packages and identifiers as `cffconvert` and other CFF ecosystem tools has benefits
	yamlWorkhorse = ruamel.yaml.YAML()

	def srsly(Z0Z_field: Any, Z0Z_value: Any) -> bool:
		if Z0Z_value: # empty lists
			return True
		else:
			return False

	# Convert the attrs object to a dictionary.
	dictionaryCitation = attrs.asdict(nexusCitation, filter=srsly)

	# Rename the keys to match the CFF format.
	for keyName in list(dictionaryCitation.keys()):
		dictionaryCitation[keyName.replace("DASH", "-")] = dictionaryCitation.pop(keyName)

	# This function and this context manager only exist to work around the fact that `ruamel.yaml` does not support `pathlib.Path` objects.
	def writeStream(pathFilename: pathlib.Path) -> None:
		pathFilename = pathlib.Path(pathFilename)
		pathFilename.parent.mkdir(parents=True, exist_ok=True)
		with open(pathFilename, 'w') as pathlibIsAStealthContextManagerThatRuamelCannotDetectAndRefusesToWorkWith:
			yamlWorkhorse.dump(dictionaryCitation, pathlibIsAStealthContextManagerThatRuamelCannotDetectAndRefusesToWorkWith)

	# Write the validation file because I haven't figured out how to validate it as a stream yet.
	pathFilenameForValidation = pathlib.Path(pathFilenameCitationSSOT).with_stem('validation')
	writeStream(pathFilenameForValidation)
	citationObject: cffconvert.Citation = create_citation(infile=str(pathFilenameForValidation), url=None)
	pathFilenameForValidation.unlink()

	# If the validation succeeds, write the CFF file and the CFF repo file.
	if citationObject.validate() is None:
		writeStream(pathFilenameCitationSSOT)
		if pathFilenameCitationDOTcffRepo is not None and pathFilenameCitationDOTcffRepo != pathFilenameCitationSSOT:
			writeStream(pathFilenameCitationDOTcffRepo)
		return True

	return False



import pytest
from tests.conftest import standardizedEqualTo
from updateCitation.pypi import getPyPIrelease

@pytest.mark.parametrize("nexusCitationTesting, expected", [
({"title": "numpy", "version": "1.23.5"}, {"repositoryDASHartifact": "https://pypi.org/project/numpy/1.23.5/"}),
({"title": "scikit-learn", "version": "1.2.2"}, {"repositoryDASHartifact": "https://pypi.org/project/scikit-learn/1.2.2/"}),
({"title": "Requests", "version": "2.28.2"}, {"repositoryDASHartifact": "https://pypi.org/project/requests/2.28.2/"}),
({"title": "Orion", "version": "5.7.11"}, {"repositoryDASHartifact": "https://pypi.org/project/orion/5.7.11/"}),
({"title": "Taurus", "version": "11.13.17"}, {"repositoryDASHartifact": "https://pypi.org/project/taurus/11.13.17/"}),
({"version": "1.0.0"}, ValueError),
({"title": "pandas"}, ValueError),
({}, ValueError), ], indirect=["nexusCitationTesting"] )
def test_getPyPIrelease(nexusCitationTesting: dict[str, str], expected: dict[str, str] | type[ValueError]) -> None:
	"""Test PyPI release info retrieval with various package names and versions."""
	standardizedEqualTo(expected, getPyPIrelease, nexusCitationTesting)



from tests.conftest import standardizedEqualTo
from typing import Any
from unittest.mock import MagicMock, patch
from updateCitation import addGitHubRelease, addGitHubSettings, CitationNexus, SettingsPackage
from updateCitation.github import getGitHubRelease

def test_addGitHubSettings_preservesGitUserEmail(settingsPackageTesting: SettingsPackage) -> None:
	emailBefore = settingsPackageTesting.gitUserEmail
	updatedPackage = addGitHubSettings(settingsPackageTesting)
	assert updatedPackage.gitUserEmail == emailBefore, (
		f"Expected email to remain {emailBefore}, "
		f"but got {updatedPackage.gitUserEmail}"
	)

def test_getGitHubRelease_noRepository(nexusCitationTesting: CitationNexus, settingsPackageTesting: SettingsPackage) -> None:
	nexusCitationTesting.repository = None
	standardizedEqualTo(None, getGitHubRelease, nexusCitationTesting, settingsPackageTesting)

def test_addGitHubRelease_hypotheticalVersion(nexusCitationTesting: CitationNexus, settingsPackageTesting: SettingsPackage) -> None:
	nexusCitationTesting.repository = "dummyRepo"
	nexusCitationTesting.version = "9.9.9"

	with patch('updateCitation.github.getGitHubRelease') as mockGetRelease:
		mockGetRelease.return_value = None
		updatedCitation = addGitHubRelease(nexusCitationTesting, settingsPackageTesting)

	# For now, we only check that it did not throw, and returns a CitationNexus.
	assert isinstance(updatedCitation, CitationNexus), (
		"Expected addGitHubRelease to return a CitationNexus"
	)

@patch('updateCitation.github.GitHubRepository')
def test_getGitHubRelease_successfulResponse(mockGitHubRepo: MagicMock, nexusCitationTesting: CitationNexus, settingsPackageTesting: SettingsPackage) -> None:
	nexusCitationTesting.repository = "owner/repo"
	nexusCitationTesting.version = "1.0.0"

	# Mock the GitHub repository and release objects
	mockRelease = MagicMock()
	mockRelease.tag_name = "1.0.0"
	mockRelease.html_url = "https://github.com/owner/repo/releases/tag/1.0.0"
	mockRelease.published_at.strftime.return_value = "2025-06-02"

	mockRepo = MagicMock()
	mockRepo.get_latest_release.return_value = mockRelease

	mockTagRef = MagicMock()
	mockTagRef.object.sha = "abc123"
	mockTagRef.object.type = "commit"
	mockRepo.get_git_ref.return_value = mockTagRef

	mockGitHubRepo.return_value.__enter__.return_value = mockRepo

	releaseData = getGitHubRelease(nexusCitationTesting, settingsPackageTesting)

	assert releaseData is not None
	assert releaseData["commit"] == "abc123"
	assert releaseData["dateDASHreleased"] == "2025-06-02"
	assert len(releaseData["identifiers"]) == 1
	assert releaseData["identifiers"][0]["value"] == "https://github.com/owner/repo/releases/tag/1.0.0"

@patch('updateCitation.github.getGitHubRelease')
def test_addGitHubRelease_withValidReleaseData(mockGetRelease: MagicMock, nexusCitationTesting: CitationNexus, settingsPackageTesting: SettingsPackage) -> None:
	nexusCitationTesting.repository = "owner/repo"
	nexusCitationTesting.version = "1.0.0"

	mockReleaseData: dict[str, Any] = {
		"commit": "abc123",
		"dateDASHreleased": "2025-06-02",
		"identifiers": [{"type": "url", "value": "https://github.com/owner/repo/releases/tag/1.0.0"}],
		"repositoryDASHcode": "https://github.com/owner/repo/releases/tag/1.0.0"
	}
	mockGetRelease.return_value = mockReleaseData

	updatedCitation = addGitHubRelease(nexusCitationTesting, settingsPackageTesting)

	assert updatedCitation.commit == "abc123"
	assert updatedCitation.dateDASHreleased == "2025-06-02"
	assert updatedCitation.identifiers == mockReleaseData["identifiers"]
	assert updatedCitation.repositoryDASHcode == "https://github.com/owner/repo/releases/tag/1.0.0"



from pathlib import Path
from tomllib import loads as tomllib_loads
from typing import Any
from updateCitation import CitationNexus, mapNexusCitation2pyprojectDOTtoml, SettingsPackage

def getSettingsPackage(pathFilename: Path) -> SettingsPackage:
	Z0Z_tomlSherpa = tomllib_loads(pathFilename.read_text(encoding="utf-8"))
	Z0Z_SettingsPackage: dict[str, Any] = {}
	if Z0Z_tomlSherpa.get("tool", None):
		Z0Z_SettingsPackage = Z0Z_tomlSherpa["tool"].get("updateCitation", {})
	truth = SettingsPackage(**Z0Z_SettingsPackage, pathFilenamePackageSSOT=pathFilename)
	truth = get_pyprojectDOTtoml(truth)
	return truth

def get_pyprojectDOTtoml(truth: SettingsPackage) -> SettingsPackage:
	truth.tomlPackageData = tomllib_loads(truth.pathFilenamePackageSSOT.read_text(encoding="utf-8"))['project']
	return truth

def add_pyprojectDOTtoml(nexusCitation: CitationNexus, truth: SettingsPackage) -> CitationNexus:
	def Z0Z_ImaNotValidatingNoNames(person: dict[str, str]) -> dict[str, str]:
		cffPerson: dict[str, str] = {}
		if person.get('name', None):
			cffPerson['given-names'], cffPerson['family-names'] = person['name'].split(' ', 1)
		if person.get('email', None):
			cffPerson['email'] = person['email']
		return cffPerson

	packageName = truth.tomlPackageData.get("name", None)
	nexusCitation.title = packageName

	for keyNexusCitation, key_pyprojectDOTtoml in mapNexusCitation2pyprojectDOTtoml:
		listPersonsTOML = truth.tomlPackageData.get(key_pyprojectDOTtoml, None)
		if listPersonsTOML:
			listPersonsCFF: list[dict[str, str]] = []
			for person in listPersonsTOML:
				listPersonsCFF.append(Z0Z_ImaNotValidatingNoNames(person))
			setattr(nexusCitation, keyNexusCitation, listPersonsCFF)

	nexusCitation.setInStone("pyprojectDOTtoml")
	return nexusCitation



# import os
# from updateCitation import flowControl
# import pathlib
# import pytest
# from tests.conftest import standardizedEqualTo, cffDASHversionDefaultHARDCODED, messageDefaultHARDCODED, CitationNexus, SettingsPackage, get_pyprojectDOTtoml

# # Dummy implementations to record call order.
# def dummy_add_pyprojectDOTtoml(nexusCitation: CitationNexus, truth: SettingsPackage) -> tuple[CitationNexus, SettingsPackage]:
# 	callOrder.append("add_pyprojectDOTtoml")
# 	truth = get_pyprojectDOTtoml(truth)  # Populate truth.tomlPackageData from the pyproject.toml file.
# 	# Set required fields to avoid ValueError from setInStone.
# 	nexusCitation.title = truth.tomlPackageData.get("name")
# 	nexusCitation.authors = truth.tomlPackageData.get("authors", [])
# 	# Supply a dummy contact if absent.
# 	contactData = truth.tomlPackageData.get("contact")
# 	if not contactData:
# 		nexusCitation.contact = [{"name": "Dummy", "email": "dummy@example.com"}]
# 	else:
# 		nexusCitation.contact = contactData
# 	return nexusCitation, truth

# def dummy_addCitation(nexusCitation: CitationNexus, pathFilenameCitationSSOT) -> CitationNexus:
# 	callOrder.append("addCitation")
# 	return nexusCitation

# def dummy_addPyPAMetadata(nexusCitation: CitationNexus, tomlPackageData: dict) -> CitationNexus:
# 	callOrder.append("addPyPAMetadata")
# 	return nexusCitation

# def dummy_addGitHubSettings(truth: SettingsPackage) -> SettingsPackage:
# 	callOrder.append("addGitHubSettings")
# 	return truth

# def dummy_addGitHubRelease(nexusCitation: CitationNexus, truth: SettingsPackage) -> CitationNexus:
# 	callOrder.append("addGitHubRelease")
# 	return nexusCitation

# def dummy_addPyPIrelease(nexusCitation: CitationNexus) -> CitationNexus:
# 	callOrder.append("addPyPIrelease")
# 	return nexusCitation

# def dummy_writeCitation(nexusCitation: CitationNexus, pathFilenameCitationSSOT, pathFilenameCitationDOTcffRepository) -> bool:
# 	callOrder.append("writeCitation")
# 	return True

# def dummy_gittyUpGitPushGitHub(truth: SettingsPackage, nexusCitation: CitationNexus, pathFilenameCitationSSOT, pathFilenameCitationDOTcffRepository) -> bool:
# 	callOrder.append("gittyUpGitPushGitHub")
# 	return True

# @pytest.fixture
# def validTomlTmpFile(tmp_path: pathlib.Path) -> pathlib.Path:
# 	pathFilename = tmp_path / "pyproject.toml"
# 	tomlContent = """
# [project]
# name = "ValidTitle"
# authors = [{name = "John Doe", email = "john@example.com"}]
# contact = [{name = "Contact Name", email = "contact@example.com"}]
# """
# 	pathFilename.write_text(tomlContent)
# 	return pathFilename

# @pytest.fixture
# def missingTitleTomlTmpFile(tmp_path: pathlib.Path) -> pathlib.Path:
# 	pathFilename = tmp_path / "pyproject.toml"
# 	# Missing the "name" key.
# 	tomlContent = """
# [project]
# authors = [{name = "John Doe", email = "john@example.com"}]
# contact = [{name = "Contact Name", email = "contact@example.com"}]
# """
# 	pathFilename.write_text(tomlContent)
# 	return pathFilename

# def restore_flowControl(monkeypatch) -> None:
# 	# Restore original functions (if needed in further tests) by deleting our attributes.
# 	for name in ["add_pyprojectDOTtoml", "addCitation", "addPyPAMetadata", "addGitHubRelease", "addPyPIrelease", "writeCitation"]:
# 		if hasattr(flowControl, name):
# 			monkeypatch.undo()

# def test_here_valid(validTomlTmpFile: pathlib.Path, monkeypatch: pytest.MonkeyPatch) -> None:
# 	global callOrder
# 	callOrder = []  # type: list[str]

# 	# Override functions in flowControl, recording call order.
# 	monkeypatch.setattr(flowControl, "add_pyprojectDOTtoml", dummy_add_pyprojectDOTtoml)
# 	monkeypatch.setattr(flowControl, "addCitation", dummy_addCitation)
# 	monkeypatch.setattr(flowControl, "addPyPAMetadata", dummy_addPyPAMetadata)
# 	monkeypatch.setattr(flowControl, "addGitHubSettings", dummy_addGitHubSettings)
# 	monkeypatch.setattr(flowControl, "addGitHubRelease", dummy_addGitHubRelease)
# 	monkeypatch.setattr(flowControl, "addPyPIrelease", dummy_addPyPIrelease)
# 	monkeypatch.setattr(flowControl, "writeCitation", dummy_writeCitation)
# 	monkeypatch.setattr(flowControl, "gittyUpGitPushGitHub", dummy_gittyUpGitPushGitHub)

# 	# Call here() with our valid temporary pyproject.toml.
# 	flowControl.here(validTomlTmpFile)

# 	# Verify call order.
# 	expectedOrder = [
# 		"add_pyprojectDOTtoml",
# 		"addCitation",
# 		"addPyPAMetadata",
# 		"addGitHubSettings",
# 		"addGitHubRelease",
# 		"addPyPIrelease",
# 		"writeCitation",
# 		"gittyUpGitPushGitHub"
# 	]
# 	assert callOrder == expectedOrder

# def test_here_missingTitle(missingTitleTomlTmpFile: pathlib.Path, monkeypatch: pytest.MonkeyPatch) -> None:
# 	# Override add_pyprojectDOTtoml to mimic its normal behavior so that title remains None.
# 	def dummy_noTitle(nexusCitation: CitationNexus, truth: SettingsPackage) -> tuple[CitationNexus, SettingsPackage]:
# 		# Do not set title.
# 		nexusCitation.authors = truth.tomlPackageData.get("authors", [])
# 		nexusCitation.contact = truth.tomlPackageData.get("contact", [])
# 		# This will cause setInStone to raise ValueError.
# 		nexusCitation.setInStone("pyprojectDOTtoml")
# 		return nexusCitation, truth

# 	monkeypatch.setattr(flowControl, "add_pyprojectDOTtoml", dummy_noTitle)

# 	with pytest.raises(ValueError):
# 		flowControl.here(missingTitleTomlTmpFile)



from typing import Any
from updateCitation import CitationNexus
import packaging.utils

def getPyPIrelease(nexusCitation: CitationNexus) -> dict[str, Any]:
	if not nexusCitation.title:
		raise ValueError("Package name (title) is required to get PyPI release info.")
	if not nexusCitation.version:
		raise ValueError("Package version is required to get PyPI release info.")

	packageName = packaging.utils.canonicalize_name(nexusCitation.title)
	version = str(nexusCitation.version)
	return {"repositoryDASHartifact": f"https://pypi.org/project/{packageName}/{version}/"}

def addPyPIrelease(nexusCitation: CitationNexus) -> CitationNexus:
	pypiReleaseData = getPyPIrelease(nexusCitation)
	nexusCitation.repositoryDASHartifact = pypiReleaseData.get("repositoryDASHartifact")

	nexusCitation.setInStone("PyPI")
	return nexusCitation
